﻿#region License
/*
SecondLanguage Gettext Library for .NET
Copyright 2013 James F. Bellinger <http://www.zer7.com>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
not claim that you wrote the original software. If you use this software
in a product, an acknowledgement in the product documentation would be
appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must
not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
distribution.
*/
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NUnit.Framework;

namespace SecondLanguage.Tests
{
    [TestFixture]
    sealed class GettextMOTranslationTest
    {
        [Test]
        public void Headers()
        {
            var a = new GettextMOTranslation();
            Assert.AreEqual("8bit", a.GetHeader("Content-Transfer-Encoding"));
            Assert.AreEqual("text/plain; charset=utf-8", a.GetHeader("Content-Type"));
            Assert.AreEqual(Encoding.UTF8, a.Encoding);
            
            a.SetHeader("Content-Type", "text/plain; charset=ascii");
            Assert.AreEqual(Encoding.ASCII, a.Encoding);
        }

        [Test]
        public void Simple()
        {
            var a = new GettextMOTranslation();
            a.SetString("A", "B");
            Assert.AreEqual("B", a.GetString("A"));
        }

        [Test]
        public void ReadOnly()
        {
            var a = new GettextMOTranslation();
            a.SetString("A", "B");
            a.MakeReadOnly();
            Assert.Throws<InvalidOperationException>(() => a.Clear());
            Assert.Throws<InvalidOperationException>(() => a.SetHeader("Q", "R"));
            Assert.Throws<InvalidOperationException>(() => a.SetPluralString("A", "B", new[] { "C", "D" }));
            Assert.Throws<InvalidOperationException>(() => a.SetString("A", "B"));
            Assert.Throws<InvalidOperationException>(() => a.IsLittleEndian = true);
        }

        [Test]
        public void SaveLoadContextToDisk()
        {
            var a = new GettextMOTranslation();
            a.SetString("A", "B", "Hello");
            a.SetString("A", "C", "Goodbye");
            a.SetString("A", "D");
            a.Save("test.mo");

            var b = new GettextMOTranslation();
            b.Load("test.mo");
            Assert.AreEqual("B", b.GetString("A", "Hello"));
            Assert.AreEqual("C", b.GetString("A", "Goodbye"));
            Assert.AreEqual("D", b.GetString("A"));
        }

        [Test]
        public void Plural()
        {
            var a = new GettextMOTranslation();
            a.SetPluralString("Singular", "Plural", new[] { "One", "Not One" });
            Assert.AreEqual("One", a.GetPluralStringByIndex("Singular", "Plural", 0));
            Assert.AreEqual("Not One", a.GetPluralStringByIndex("Singular", "Plural", 1));
            Assert.AreEqual("Not One", a.GetPluralStringByIndex("Singular", "Plural", a.GetPluralIndex(0)));
            Assert.AreEqual("One", a.GetPluralStringByIndex("Singular", "Plural", a.GetPluralIndex(1)));
            Assert.AreEqual("Not One", a.GetPluralStringByIndex("Singular", "Plural", a.GetPluralIndex(2)));
        }

        [Test]
        public void SaveLoadPlural()
        {
            var a = new GettextMOTranslation();
            a.SetPluralString("A", "B", new[] { "X", "Y" });
            var buffer = a.Save();

            var b = new GettextMOTranslation();
            b.Load(buffer);
            Assert.AreEqual("X", b.GetPluralStringByIndex("A", "B", 0));
            Assert.AreEqual("Y", b.GetPluralStringByIndex("A", "B", 1));
        }

        [Test]
        public void SaveLoadPluralToDisk()
        {
            var a = new GettextMOTranslation();
            a.SetPluralString("A", "B", new[] { "X", "Y" });
            a.Save("test.mo");

            var b = new GettextMOTranslation();
            b.Load("test.mo");
            Assert.AreEqual("X", b.GetPluralStringByIndex("A", "B", 0));
            Assert.AreEqual("Y", b.GetPluralStringByIndex("A", "B", 1));
        }

        [Test]
        public void PluralForms()
        {
            var a = new GettextMOTranslation();

            // See http://www.gnu.org/software/gettext/manual/html_node/Plural-forms.html for the following plural form.
            // Its complexity should work out the Plural-Forms parser rather well...
            a.SetHeader("Plural-Forms", "nplurals=3; plural=n==1 ? 0 : n%10>=2 && n%10<=4 && (n%100<10 || n%100>=20) ? 1 : 2");
            Assert.AreEqual(0, a.GetPluralIndex(1));
            Assert.AreEqual(1, a.GetPluralIndex(2));
            Assert.AreEqual(2, a.GetPluralIndex(21));
            Assert.AreEqual(2, a.GetPluralIndex(14));
            Assert.AreEqual(1, a.GetPluralIndex(24));
            Assert.AreEqual(2, a.GetPluralIndex(15));
            Assert.AreEqual(1, a.GetPluralIndex(132));
            Assert.AreEqual(2, a.GetPluralIndex(112));
        }

        [Test]
        public void GetGettextKeys()
        {
            var a = new GettextMOTranslation();
            a.SetString("A", "X");
            a.SetString("B", "Y");
            a.SetString("C", "Z");

            var d = new Dictionary<GettextKey, string[]>();
            foreach (var kvp in a.GetGettextKeys()) { d[kvp.Key] = kvp.Value; }

            Assert.AreEqual(4, d.Count);
            Assert.AreEqual("X", d["A"][0]);
            Assert.AreEqual("Y", d["B"][0]);
            Assert.AreEqual("Z", d["C"][0]);
        }
    }
}
