﻿#region License
/*
SecondLanguage Gettext Library for .NET
Copyright 2013 James F. Bellinger <http://www.zer7.com>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
not claim that you wrote the original software. If you use this software
in a product, an acknowledgement in the product documentation would be
appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must
not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
distribution.
*/
#endregion

using System.Linq;
using System.Text;
using NUnit.Framework;

namespace SecondLanguage.Tests
{
    [TestFixture]
    sealed class GettextPOTranslationTest
    {
        [Test]
        public void Headers()
        {
            var a = new GettextPOTranslation();
            Assert.AreEqual("8bit", a.GetHeader("Content-Transfer-Encoding"));
            Assert.AreEqual("text/plain; charset=utf-8", a.GetHeader("Content-Type"));
            Assert.AreEqual(Encoding.UTF8, a.Encoding);

            a.SetHeader("Content-Type", "text/plain; charset=ascii");
            Assert.AreEqual(Encoding.ASCII, a.Encoding);
        }

        [Test]
        public void SaveLoadDefaults()
        {
            var a = new GettextPOTranslation();
            var saved = a.Save();

            var b = new GettextPOTranslation();
            b.Load(saved);

            Assert.AreEqual(a.ToString(), b.ToString());
        }

        [Test]
        public void SaveLoadContexts()
        {
            var a = new GettextPOTranslation();
            a.SetString("A", "B", "C");
            a.SetString("A", "D");
            var saved = a.Save();

            var b = new GettextPOTranslation();
            b.Load(saved);

            Assert.AreEqual("B", a.GetString("A", "C"));
            Assert.AreEqual("D", a.GetString("A"));
        }

        [Test]
        public void SaveLoadPlural()
        {
            var a = new GettextPOTranslation();
            a.SetPluralString("A", "B", new[] { "X", "Y" }, "C");
            var saved = a.Save();

            var b = new GettextPOTranslation();
            b.Load(saved);

            Assert.AreEqual("X", a.GetPluralString("A", "B", 1, "C"));
            Assert.AreEqual("Y", a.GetPluralString("A", "B", 0, "C"));
            Assert.AreEqual(null, a.GetPluralString("A", "B", 1));
            Assert.AreEqual(null, a.GetPluralString("A", "B", 0));

            b.SetPluralString("A", "B", new[] { "U", "V" }, "C");

            Assert.AreEqual("U", b.GetPluralString("A", "B", 1, "C"));
            Assert.AreEqual("V", b.GetPluralString("A", "B", 0, "C"));
        }

        [Test]
        public void LossyASCII()
        {
            var a = new GettextPOTranslation();
            a.SetString("T", "θ");
            a.Encoding = Encoding.ASCII;
            var saved = a.Save();

            var b = new GettextPOTranslation();
            b.Load(saved);

            Assert.AreEqual("?", b.GetString("T"));
        }

        [Test]
        public void GettextKeysAreKeptInAddOrderInPOFormat()
        {
            var a = new GettextPOTranslation();
            a.SetString("F", "Z");
            a.SetString("E", "Y");
            a.SetString("D", "X");
            a.SetString("C", "W");
            a.SetString("B", "V");
            a.SetString("A", "U");

            Assert.AreEqual(new[] { "", "F", "E", "D", "C", "B", "A" },
                            a.GetGettextKeys().Select(x => x.Key.ID).ToArray());

            Assert.AreEqual(new[] { "Z", "Y", "X", "W", "V", "U" },
                            a.GetGettextKeys().Where(x => x.Key.ID != "").Select(x => x.Value[0]).ToArray());
        }
    }
}
