﻿#region License
/*
SecondLanguage Gettext Library for .NET
Copyright 2013 James F. Bellinger <http://www.zer7.com>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
not claim that you wrote the original software. If you use this software
in a product, an acknowledgement in the product documentation would be
appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must
not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
distribution.
*/
#endregion

using System;
using System.Globalization;
using System.IO;
using NUnit.Framework;

namespace SecondLanguage.Tests
{
    [TestFixture]
    class TranslatorTest
    {
        [Test]
        public void Default()
        {
            Assert.AreEqual(Translator.Default.Translate("Test"), "Test");
        }

        [Test]
        public void NoFiles()
        {
            var t = new Translator();
            Assert.AreEqual(t.Translate("Test"), "Test");
            Assert.AreEqual(t.TranslatePlural("Singular", "Plural", 0), "Plural");
            Assert.AreEqual(t.TranslatePlural("Singular", "Plural", 1), "Singular");
            Assert.AreEqual(t.TranslatePlural("Singular", "Plural", 2), "Plural");
            Assert.AreEqual(t.TranslatePlural("Singular {0}", "Plural {0}", 2, 42), "Plural 42");
        }

        [Test]
        public void Fallbacks()
        {
            var a = new GettextMOTranslation();
            var b = new GettextMOTranslation();
            var t = new Translator();
            t.RegisterTranslation(a);
            t.RegisterTranslation(b);

            a.SetString("Test", "Hello");
            Assert.AreEqual("Hello", t["Test"]);
            Assert.AreEqual("Toast", t["Toast"]);

            b.SetString("Toast", "Tasty");
            Assert.AreEqual("Tasty", t["Toast"]);

            a.SetString("Toast", "Burnt");
            Assert.AreEqual("Burnt", t["Toast"]);
        }

        [Test]
        public void ReadOnly()
        {
            var t = new Translator();
            t.MakeReadOnly();
            Assert.Throws<InvalidOperationException>(() => t.Parent = null);
            Assert.Throws<InvalidOperationException>(() => t.RegisterTranslation(new GettextMOTranslation()));
        }

        [Test]
        public void ThrowOnFormatExceptions()
        {
            var a = new GettextMOTranslation();
            var b = new GettextMOTranslation();
            var t = new Translator();
            t.RegisterTranslation(a);
            t.RegisterTranslation(b);

            b.SetString("Test {1}", "Hello {2}");
            Assert.AreEqual("Test {1}", t["Test {1}", 42]);
            Assert.AreEqual("Test 43", t["Test {1}", 42, 43]);
            Assert.AreEqual("Hello 44", t["Test {1}", 42, 43, 44]);

            a.SetString("Test {1}", "Goodbye {2}");
            Assert.AreEqual("Goodbye 44", t.Translate("Test {1}", 42, 43, 44));

            t.ThrowOnFormatExceptions = true;
            Assert.Throws<FormatException>(() => t.Translate("Test {1}", 42, 43));
            Assert.AreEqual("Goodbye 44", t.Translate("Test {1}", 42, 43, 44));
        }

        [Test]
        public void RegisterTranslationsByCultureDefaultDirectory()
        {
            var mo = new GettextPOTranslation();
            mo.SetString("A", "B");
            mo.Save(Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "hello.po"));

            var t = new Translator();
            t.RegisterTranslationsBySearchPattern("*.po");
            Assert.AreEqual("B", t["A"]);
        }

        [Test]
        public void RegisterTranslationsByCulture()
        {
            string prefix = Path.Combine(Path.GetTempPath(), "test-locale");

            var t = new Translator();
            try { Directory.CreateDirectory(Path.Combine(prefix, "en")); } catch { }
            try { Directory.CreateDirectory(Path.Combine(prefix, "en-US")); } catch { }
            try { Directory.CreateDirectory(Path.Combine(prefix, "ja")); } catch { }
            File.Delete(Path.Combine(prefix, @"en\test.mo"));
            File.Delete(Path.Combine(prefix, @"en-US\test.mo"));
            File.Delete(Path.Combine(prefix, @"ja\test.mo"));

            var a = new GettextMOTranslation();
            a.SetString("A", "B");
            a.Save(Path.Combine(prefix, @"ja\test.mo"));
            
            t.ClearTranslationList();
            t.RegisterTranslationsByCulture(new[] { @"{0}\test.mo" }, new[] { new CultureInfo("en-US"), new CultureInfo("ja-JP") }, prefix);
            Assert.AreEqual("B", t.Translate("A"));

            var b = new GettextMOTranslation();
            b.SetString("A", "C");
            b.Save(Path.Combine(prefix, @"en-US\test.mo"));

            t.ClearTranslationList();
            t.RegisterTranslationsByCulture(new[] { @"{0}\test.mo" }, new[] { new CultureInfo("en-US"), new CultureInfo("ja-JP") }, prefix);
            Assert.AreEqual("C", t.Translate("A"));

            var c = new GettextMOTranslation();
            c.SetString("A", "D");
            c.Save(Path.Combine(prefix, @"en\test.mo"));

            t.ClearTranslationList();
            t.RegisterTranslationsByCulture(new[] { @"{0}\test.mo" }, new[] { new CultureInfo("en-US"), new CultureInfo("ja-JP") }, prefix);
            Assert.AreEqual("C", t["A"]);
        }
    }
}
